@@ -231,11 +231,7 @@ static char *(features[]) =
 #else
 	"-dnd",
 #endif
-#ifdef EBCDIC
-	"+ebcdic",
-#else
 	"-ebcdic",
-#endif
 #ifdef FEAT_EMACS_TAGS
 	"+emacs_tags",
 #else
@@ -750,6 +746,190 @@ static char *(features[]) =
 
 static int included_patches[] =
 {   /* Add new patch number below this line */
+/**/
+    4328,
+/**/
+    4327,
+/**/
+    4326,
+/**/
+    4325,
+/**/
+    4324,
+/**/
+    4323,
+/**/
+    4322,
+/**/
+    4321,
+/**/
+    4320,
+/**/
+    4319,
+/**/
+    4318,
+/**/
+    4317,
+/**/
+    4316,
+/**/
+    4315,
+/**/
+    4314,
+/**/
+    4313,
+/**/
+    4312,
+/**/
+    4311,
+/**/
+    4310,
+/**/
+    4309,
+/**/
+    4308,
+/**/
+    4307,
+/**/
+    4306,
+/**/
+    4305,
+/**/
+    4304,
+/**/
+    4303,
+/**/
+    4302,
+/**/
+    4301,
+/**/
+    4300,
+/**/
+    4299,
+/**/
+    4298,
+/**/
+    4297,
+/**/
+    4296,
+/**/
+    4295,
+/**/
+    4294,
+/**/
+    4293,
+/**/
+    4292,
+/**/
+    4291,
+/**/
+    4290,
+/**/
+    4289,
+/**/
+    4288,
+/**/
+    4287,
+/**/
+    4286,
+/**/
+    4285,
+/**/
+    4284,
+/**/
+    4283,
+/**/
+    4282,
+/**/
+    4281,
+/**/
+    4280,
+/**/
+    4279,
+/**/
+    4278,
+/**/
+    4277,
+/**/
+    4276,
+/**/
+    4275,
+/**/
+    4274,
+/**/
+    4273,
+/**/
+    4272,
+/**/
+    4271,
+/**/
+    4270,
+/**/
+    4269,
+/**/
+    4268,
+/**/
+    4267,
+/**/
+    4266,
+/**/
+    4265,
+/**/
+    4264,
+/**/
+    4263,
+/**/
+    4262,
+/**/
+    4261,
+/**/
+    4260,
+/**/
+    4259,
+/**/
+    4258,
+/**/
+    4257,
+/**/
+    4256,
+/**/
+    4255,
+/**/
+    4254,
+/**/
+    4253,
+/**/
+    4252,
+/**/
+    4251,
+/**/
+    4250,
+/**/
+    4249,
+/**/
+    4248,
+/**/
+    4247,
+/**/
+    4246,
+/**/
+    4245,
+/**/
+    4244,
+/**/
+    4243,
+/**/
+    4242,
+/**/
+    4241,
+/**/
+    4240,
+/**/
+    4239,
+/**/
+    4238,
+/**/
+    4237,
 /**/
     4236,
 /**/
@@ -9368,7 +9548,7 @@ list_in_columns(char_u **items, int size, int current)
     // The rightmost column doesn't need a separator.
     // Sacrifice it to fit in one more column if possible.
     ncol = (int) (Columns + 1) / width;
-    nrow = item_count / ncol + (item_count % ncol ? 1 : 0);
+    nrow = item_count / ncol + ((item_count % ncol) ? 1 : 0);
 
     // "i" counts columns then rows.  "idx" counts rows then columns.
     for (i = 0; !got_int && i < nrow * ncol; ++i)
